#!/usr/sbin/perl -w
#
# ReverseMRP.pl v1.0a
# Last modified April 29, 2003 11:11
# (c) Olaf R.P. Bininda-Emonds
#
# Expects / requires:
#   A copy of paup4b10 in the working path
#   A nexus-formatted data file with charsets corresponding to the
#      different source trees
#
# Generates:
#   A nexus-formatted tree file containing the source tree descriptions
#   corresponding to their matrix representations
#
# Usage: ReverseMRP.pl -f<filename> [-m] [-p] [-s] [-v] [-h]
#   options: -f = name of NEXUS-formatted data file
#            -m = output pruned data matrix for each source tree to log file
#            -p = purge temporary files
#            -s = perform search for each source tree
#            -v = verbose output
#            -h = print this message and quit

$filename = "";
$matrix = 0;
$purge = 0;
$search = 0;
$verbose = 0;
$charset = 0;

for ($i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-f([\w+\.?\w+?]*)/)
		{
		$filename = $1;
		}
	elsif ($ARGV[$i] =~ /-m/)
		{
		$matrix = 1;
		}
	elsif ($ARGV[$i] =~ /-p/)
		{
		$purge = 1;
		}
	elsif ($ARGV[$i] =~ /-s/)
		{
		$search = 1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: ReverseMRP.pl -f<filename> [-m] [-s] [-v] [-h]\n";
		print "Options: -f = name of NEXUS-formatted data file\n";
		print "         -m = output pruned data matrix for each source tree to log file\n";
		print "         -p = purge temporary files\n";
		print "         -s = perform search for each source tree\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: ReverseMRP.pl -f<filename> [-m] [-p] [-s] [-v] [-h]\n";
		exit(1); 
		}
	}

die "Must supply name of nexus-formatted file containing MRP matrix\n" if (not $filename);

setLineBreak($filename);
open (fDATA, "<$filename") or die "Cannot open MRP datafile $filename\n";		#Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	$readFlag = $start = $stop = 0;
	$taxon = $data = $study = $chars = "";
	undef @taxonList;
	undef @dataList;
	undef @studyList;
	undef @startChar;
	undef @stopChar;
	while (<fDATA>)
		{
		#chomp;
		$line = $_;
		$line =~ s/^\s+//;		#Strip (any) leading white space
		if ($line =~ m/matrix/i)
			{
			$readFlag = 1;
			}
		if ($readFlag)
			{
			if ($_ =~ m/;/)
				{
				$readFlag = 0;
				}
			unless ($line =~ m/matrix/i or $line eq "" or $line =~ m/\[/ or $line =~ m/;/)
				{
				($taxon, $data) = split(/\s+/, $line);
				push @taxonList, $taxon;
				push @dataList, $data;
				}
			}
		if ($line =~ m/charset/i)
			{
			$charset = 1;
			chomp $line;
			($study, $chars) = split (/=/, $line);
			$study =~ s/charset\s+//i;
			$study =~ s/\s+$//;
			push @studyList, $study;
			$chars =~ s/^\s+//;
			$chars =~ s/;$//;
			if ($chars =~ m/-/)
				{
				($start, $stop) = split(/-/, $chars);
				}
			else
				{
				$start = $chars;
				$stop = $chars;
				}
			push @startChar, $start;
			push @stopChar, $stop;
#			$NumChars = $stop - $start + 1;
#			print "Charset: $start\t$stop\t$NumChars\n";
			}
		}
close fDATA;

if ($verbose)
	{
	print "Data matrix ...\n";
	for ($count = 0; $count < @taxonList; $count++)
		{
		print "\t$taxonList[$count]\t$dataList[$count]\n";
		}
	print "\nCharacter sets ...\n";
	for ($count = 0; $count < @studyList; $count++)
		{
		}
	}

die "$filename does not contain charsets corresponding to source trees\n" if (not $charset);

($outFile) = split(/\./, $filename, 0);
$outCmd = $outFile."RVS.txt";
$outTree = $outFile."RVS.tre";

open (fCMD, ">$outCmd") or die "Cannot open batch file $outCmd\n";
	print fCMD "\nbegin paup;\n\tlog file=PossibleError.log replace;\n";
	print fCMD "\tset maxtrees=100 increase=no autoclose=yes notifybeep=no tcompress=yes torder=left\n";
	print fCMD "\t\tcriterion=parsimony status=no errorbeep=no notifybeep=no outroot=monophyl;\n";

	for ($study = 0; $study < @studyList; $study++)
		{
		print "\t$studyList[$study] --\tFrom: $startChar[$study]\tTo: $stopChar[$study]\n" if ($verbose);
		print fCMD "\n\t\[\!\]\[\!Source study: $studyList[$study]\]\n";
		$taxSet = $probSet = "";
		for ($species = 0; $species < @taxonList; $species++)
			{
			$sourceData = substr($dataList[$species], $startChar[$study] - 1, $stopChar[$study] - $startChar[$study] + 1);
			unless ($sourceData =~ m/\d/)	#Flag species containing all missing data
				{
				$taxSet .= " ".$taxonList[$species];
				}
			if ($sourceData =~ m/\?/ and $sourceData =~ m/\d/)	#Flag species containing some missing data
				{
				$probSet .= " ".$taxonList[$species];
				}
			}
		if ($verbose)
			{
			print "\t\tMissing taxa: ";
			if ($taxSet)
				{
				print "$taxSet\n";
				}
			else
				{
				print "none\n";
				}
			print "\t\tProblematic taxa: ";
			if ($probSet)
				{
				print "$probSet\n";
				}
			else
				{
				print "none\n";
				}
			}
		print fCMD "\t\t\[\!\]\[\!Problematic taxa exist: $probSet\]\n" if ($probSet);
		if ($taxSet)
			{
			print fCMD "\t\tdelete$taxSet / only;\n";
			}
		else
			{
			print fCMD "\t\trestore all;\n";
			}
		print fCMD "\t\tinclude $startChar[$study] - $stopChar[$study] / only;\n";
		print fCMD "\t\t\tshowmatrix;\n" if ($matrix);
		if ($search)
			{
#			print fCMD "\t\t\tbandb;\n";
			print fCMD "\t\t\thsearch / addseq=closest;\n";
			print fCMD "\t\t\tcontree / strict=yes semi=no majrule=no adams=no treefile=treefileRVS.tre ";
			if (not $study)
				{
				print fCMD "replace;\n";
				}
			else
				{
				print fCMD "append;\n";
				}
			print fCMD "\t\t\tcleartrees nowarn=yes;\n";
			}
		}
	print fCMD "end;\n";
	print fCMD "\nlog stop;\n";
	print fCMD "\nquit warntsave=no;\n";
close fCMD;

$catString = "cat $filename $outCmd > SourceStrip.txt";
system ($catString);
print "Obtaining source trees ...\n";
system ("nohup paup4b10 SourceStrip.txt > nohup.out");

if ($search)
	{
	print "Cleaning up tree file ...\n";
	open (fTwrite, ">$outTree") or die "Cannot open output tree file $outTree\n";
		print fTwrite "#NEXUS\n\nbegin trees;\n";
	setLineBreak(treefileRVS.tre);
	open (fTread, "<treefileRVS.tre") or die "Cannot open input tree file treefileRVS.tre\n";
		$treeCount = 0;
		while (<fTread>)
			{
			chomp;
			$treeName = $tree = "";
			if ($_ =~ /tree strict =/i)
				{
				($treeName, $tree) = split(/=/);
				$tree =~ s/\[&U\] \(outgroup,//i;
				$tree =~ s/\);$/;/i;
				print fTwrite "\ttree $studyList[$treeCount] =$tree\n";
				$treeCount++;
				}
			}
		print fTwrite "end;\n";
	
	close fTread;
	close fTwrite;
	}

system ("rm SourceStrip.txt nohup.out treefileRVS.tre") if ($purge);

# Subroutines used in the script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}
